local ADDON_NAME, addonTable = ...
local ElkFactions = LibStub("AceAddon-3.0"):NewAddon("ElkFactions", "AceBucket-3.0", "AceConsole-3.0", "AceHook-3.0")

local string_format = string.format
local string_gsub   = string.gsub

local MAX_FACTION_INDEX = 200

local LDB = LibStub("LibDataBroker-1.1")
local LDBIcon = LibStub("LibDBIcon-1.0")
local LibQTip = LibStub("LibQTip-1.0")

local gender = UnitSex("player")

local reputationSessionBase = {}

local DBdefaults = {
	profile = {
		minimap = {
			hide = false,
		},
		patternChat = "{n}: {s} - {rc}/{rm} ({rp}) [{rd}]",
		patternText = "{cw}{n}{cr}: {cb}{s}{cr} - {ci}{rc}/{rm} ({rp}){cr} [{cd}{rd}{cr}]",
		tooltipColorName = "atWar",
		tooltipColorStandingText = "standingBlizzard",
		tooltipColorStandingValueAbsolute = "standingIncremential",
		tooltipColorStandingValuePercent = "standingIncremential",
		tooltipColorStandingValueDelta = "delta",
		tooltipShowHint = false,
	},
}


-- -----
-- HELPER FUNCTIONS
-- -----

local watchedFactionIndex = nil
local function GetWatchedFactionIndex()
	local watchedID = select(6, GetWatchedFactionInfo())
	if not watchedID then
		-- no faction is currently watched
		watchedFactionIndex = nil
		return nil
	end
	if watchedFactionIndex and select(14, GetFactionInfo(watchedFactionIndex)) == watchedID then
		-- we have an index cached and it's still correct
		return watchedFactionIndex
	end
	
	local factionIndex = 1
	repeat
		if select(12, GetFactionInfo(factionIndex)) then
			watchedFactionIndex = factionIndex
			return watchedFactionIndex
		end
		factionIndex = factionIndex + 1
	until factionIndex > MAX_FACTION_INDEX
end

local function SetWatchedFactionName(factionName)
--	self:Print("SetWatchedFactionName(", factionName, ") called...")
	if not factionName then
		SetWatchedFactionIndex(0)
		return
	end
	local factionIndex = GetFactionIndexByName(factionName)
	if factionIndex then
		SetWatchedFactionIndex(factionIndex)
	end
end

local function GetFactionIndexByName(factionName)
	local factionIndex = 1
	repeat
		local name = GetFactionInfo(factionIndex)
		if name == nil then break end
		if name == factionName then
			return factionIndex
		end
		factionIndex = factionIndex + 1
	until factionIndex > MAX_FACTION_INDEX
end

-- -----
local function do_OnEnter(frame)
	local tooltip = LibQTip:Acquire("ElkFactionsTip")
	tooltip:SmartAnchorTo(frame)
	tooltip:SetAutoHideDelay(0.1, frame)
	tooltip:EnableMouse(true)
	ElkFactions:UpdateTooltip()
	tooltip:Show()
end

local function do_OnLeave()
	-- empty dummy
end

local function do_OnClick(frame, button)
	if button == "RightButton" then
		InterfaceOptionsFrame_OpenToCategory("ElkFactions")
	else
		if IsShiftKeyDown() then
			local watchedIndex = GetWatchedFactionIndex()
			if watchedIndex then
				local activeWindow = ChatEdit_GetActiveWindow()
				if activeWindow then
					local text = ElkFactions:GetFactionText(ElkFactions.db.profile.patternChat, watchedIndex, false)
					activeWindow:Insert(text)
				end
			end
		else
			ToggleCharacter("ReputationFrame")
		end
	end
end

function ElkFactions:OnInitialize()
	self.dbo = LDB:NewDataObject("ElkFactions", {
			type = "data source",
			text = "",
			icon = [[Interface\Icons\Spell_Holy_SealOfSalvation]],
			OnEnter = do_OnEnter,
			OnLeave = do_OnLeave,
			OnClick = do_OnClick,
		})

	self.db = LibStub("AceDB-3.0"):New("ElkFactionsDB", DBdefaults, true)
	self.db.RegisterCallback(self, "OnProfileChanged", "RefreshConfig")
	self.db.RegisterCallback(self, "OnProfileCopied", "RefreshConfig")
	self.db.RegisterCallback(self, "OnProfileReset", "RefreshConfig")

	LDBIcon:Register("ElkFactions", self.dbo, self.db.profile.minimap)
	
	LibStub("AceConfigDialog-3.0"):AddToBlizOptions("ElkFactions", "ElkFactions")
end

function ElkFactions:OnEnable()
	-- events for faction data changes
	self:RegisterBucketEvent({"UPDATE_FACTION"}, .1, "FactionUpdate")

	-- trying to fix ticket #9
	self:RegisterBucketEvent("PLAYER_GUILD_UPDATE", .1)

	-- events for zone changes
--~ 	self:RegisterBucketEvent({"ZONE_CHANGED", "ZONE_CHANGED_INDOORS", "ZONE_CHANGED_NEW_AREA", "MINIMAP_ZONE_CHANGED", "PLAYER_ENTERING_WORLD"}, .1, "ZoneUpdate")

	-- hook UI functions to be informed of some user actions
	self:Hook("SetWatchedFactionIndex", "Hook_SetWatchedFactionIndex", true)

	-- refresh session's base reputations
	self:UpdateBaseReputations()
end

function ElkFactions:RefreshConfig()
	-- minimap icon
	LDBIcon:Refresh("ElkFactions", self.db.profile.minimap)
	
	-- auto faction
--~ 	self:ApplyAutoZoneFaction()
	
	-- LDB text
	self:UpdateText()
	
	-- tooltip
	self:UpdateTooltip()
end

-- trying to fix ticket #9
function ElkFactions:PLAYER_GUILD_UPDATE(unit)
	if unit and unit ~= "player" then return end
	self:FactionUpdate()
end

function ElkFactions:FactionUpdate()
	-- refresh session's base reputations
	self:UpdateBaseReputations()
	
	-- update the LDB text
	self:UpdateText()
	
	-- update the tooltip
	self:UpdateTooltip()
end

function ElkFactions:UpdateBaseReputations()
	local factionIndex = 1
	repeat
		local name, description, standingID, barMin, barMax, barValue, atWarWith, canToggleAtWar, isHeader, isCollapsed, hasRep, isWatched, isChild, factionID, hasBonusRepGain, canBeLFGBonus = GetFactionInfo(factionIndex)
		if name == nil then break end
		if factionID and not reputationSessionBase[factionID] then
			reputationSessionBase[factionID] = barValue
		end
		factionIndex = factionIndex + 1
	until factionIndex > MAX_FACTION_INDEX
end

local currentZoneCached
function ElkFactions:ZoneUpdate()
	local currentZone = GetRealZoneText().." - "..GetSubZoneText()
	-- if the current zone changed, refresh the automatic zone faction
	if currentZone ~= currentZoneCached then
		self:ApplyAutoZoneFaction()
		currentZoneCached = currentZone
	end
end

function ElkFactions:UpdateText()
	local watchedIndex = GetWatchedFactionIndex()
	if watchedIndex then
		self.dbo.text = self:GetFactionText(self.db.profile.patternText, watchedIndex, true)
	else
		self.dbo.text = ""
	end
end

local FACTION_COLORS_BLIZZARD = setmetatable({}, {__index = function(self, key) local color = FACTION_BAR_COLORS[key]; local value = color and ("|cff%.2x%.2x%.2x"):format(color.r * 255, color.g * 255, color.b * 255) or nil; rawset(self, key, value); return value end })
local FACTION_COLORS_INC_old = {
	[1] = "|cff8c0000",			-- hated
	[2] = "|cffff0000",			-- hostile
	[3] = "|cffff8c00",			-- unfriendly
	[4] = "|cffbfbfbf",			-- neutral
	[5] = "|cffffffff",			-- friendly
	[6] = "|cff00ff00",			-- honored
	[7] = "|cff3f66e5",			-- reverted
	[8] = "|cff9933cc",			-- exalted
}
local FACTION_COLORS_INC = {
	[1] = "|cffff0000",			-- hated
	[2] = "|cffff8c00",			-- hostile
	[3] = "|cffffff00",			-- unfriendly
	[4] = "|cffffffff",			-- neutral
	[5] = "|cff00ff00",			-- friendly
	[6] = "|cff3f66e5",			-- honored
	[7] = "|cff9933cc",			-- reverted
	[8] = "|cffe6cc80",			-- exalted
}

local function GetFactionDeltaText(factionDelta)
	return (factionDelta == 0  and "-") or (factionDelta > 0 and "+"..factionDelta) or factionDelta
end

local function GetFactionDeltaColor(factionDelta)
	return (factionDelta == 0 and "|cffffffff") or (factionDelta > 0 and "|cff00ff00") or ("|cffff0000")
end

local function tooltip_line_OnMouseUp(frame, factionIndex, button)
	local name, description, standingID, barMin, barMax, barValue, atWarWith, canToggleAtWar, isHeader, isCollapsed, hasRep, isWatched, isChild, factionID, hasBonusRepGain, canBeLFGBonus = GetFactionInfo(factionIndex)
	if IsControlKeyDown() then
		if isHeader then
			if isCollapsed then
				ExpandFactionHeader(factionIndex)
			else
				CollapseFactionHeader(factionIndex)
			end
		else
			if IsFactionInactive(factionIndex) then
				SetFactionActive(factionIndex)
			else
				SetFactionInactive(factionIndex)
			end
		end
	elseif IsShiftKeyDown() then
		local activeWindow = ChatEdit_GetActiveWindow()
		if activeWindow then
			local text = ElkFactions:GetFactionText(ElkFactions.db.profile.patternChat, factionIndex, false)
			activeWindow:Insert(text)
		end
		return
--~ 	elseif IsAltKeyDown() then
--~ 		local curZone = GetRealZoneText()
--~ 		if self.db.profile.AutoZoneFaction[curZone] and self.db.profile.AutoZoneFaction[curZone] == name then
--~ 			self.db.profile.AutoZoneFaction[curZone] = nil
--~ 		else
--~ 			self.db.profile.AutoZoneFaction[curZone] = name
--~ 		end
--~ 		self:SetAutoFaction(curZone)
	elseif not isHeader or hasRep then
		if isWatched then
			PlaySound("igMainMenuOptionCheckBoxOff")
			SetWatchedFactionIndex(0)
		else
			PlaySound("igMainMenuOptionCheckBoxOn")
			SetWatchedFactionIndex(factionIndex)
		end
	end
end

local tooltipColors = setmetatable({
	atWar = "",
	delta = "",
	standingBlizzard = "",
	standingIncremential = "",
	standingIncrementialOld = "",
}, {__index = function(self, key) return "" end})
function ElkFactions:UpdateTooltip()
	if not LibQTip:IsAcquired("ElkFactionsTip") then
		return
	end
	local tooltip = LibQTip:Acquire("ElkFactionsTip")
	tooltip:Clear()
	tooltip:SetColumnLayout(7, "CENTER", "LEFT", "LEFT", "CENTER", "CENTER", "RIGHT", "RIGHT")

	local profile = self.db.profile
	local curLine, curColumn

	curLine = tooltip:AddHeader()
	tooltip:SetCell(curLine, 1, "|cfffed100Elk|cffffffffFactions", tooltip:GetHeaderFont(), "CENTER", 7)

-- -----
--~ 	local colortest_inc = "  "
--~ 	local colortest_inc_old = "  "
--~ 	local colortest_blizz = "  "
--~ 	for i = 1, 8 do
--~ 		colortest_inc		= colortest_inc		..FACTION_COLORS_INC[i]			..GetText("FACTION_STANDING_LABEL"..i, gender).."  "
--~ 		colortest_inc_old	= colortest_inc_old	..FACTION_COLORS_INC_old[i]		..GetText("FACTION_STANDING_LABEL"..i, gender).."  "
--~ 		colortest_blizz		= colortest_blizz	..FACTION_COLORS_BLIZZARD[i]	..GetText("FACTION_STANDING_LABEL"..i, gender).."  "
--~ 	end
--~ 	curLine = tooltip:AddLine()
--~ 	tooltip:SetCell(curLine, 1, colortest_inc, nil, "CENTER", 5)
--~ 	curLine = tooltip:AddLine()
--~ 	tooltip:SetCell(curLine, 1, colortest_inc_old, nil, "CENTER", 5)
--~ 	curLine = tooltip:AddLine()
--~ 	tooltip:SetCell(curLine, 1, colortest_blizz, nil, "CENTER", 5)
--~ 	curLine = tooltip:AddLine(" ")
-- -----

	local numFactions = GetNumFactions()
	local header_main = nil
	local header_child = nil
	local indentation
	local name, description, standingID, barMin, barMax, barValue, atWarWith, canToggleAtWar, isHeader, isCollapsed, hasRep, isWatched, isChild, factionID, hasBonusRepGain, canBeLFGBonus
	for factionIndex=1, numFactions do
		name, description, standingID, barMin, barMax, barValue, atWarWith, canToggleAtWar, isHeader, isCollapsed, hasRep, isWatched, isChild, factionID, hasBonusRepGain, canBeLFGBonus = GetFactionInfo(factionIndex)
		local standingText = GetText("FACTION_STANDING_LABEL"..standingID, gender)
		if isHeader then
			if isChild then
				header_child = name
				indentation = isCollapsed and "      |TInterface\\Buttons\\UI-PlusButton-Up:14:14|t" or "      |TInterface\\Buttons\\UI-MinusButton-Up:14:14|t"
			else
				header_child = nil
				header_main = name
				indentation = isCollapsed and "|TInterface\\Buttons\\UI-PlusButton-Up:14:14|t" or "|TInterface\\Buttons\\UI-MinusButton-Up:14:14|t"
			end
		else
			indentation = (header_main and "      " or "")..(header_child and "      " or "")
		end

		-- check if this is a friendship faction 
		local friendID, friendRep, friendMaxRep, friendName, friendText, friendTexture, friendTextLevel, friendThreshold, nextFriendThreshold = GetFriendshipReputation(factionID)
		if (friendID ~= nil) then
			standingText = friendTextLevel
			if ( nextFriendThreshold ) then
				barMin, barMax, barValue = friendThreshold, nextFriendThreshold, friendRep
			else
				barMin, barMax, barValue = 0, 8400, 8400
			end
		end

		local repIs = barValue - barMin
		local repMax = barMax - barMin
		local repDelta = not friendID and reputationSessionBase[factionID] and (barValue - reputationSessionBase[factionID]) or 0 --TODO: fix for friends
		
		tooltipColors.atWar = atWarWith and "|cffff0000" or canToggleAtWar and "|cffffff00" or "|cff00ff00"
		tooltipColors.delta = GetFactionDeltaColor(repDelta)
		tooltipColors.standingBlizzard = FACTION_COLORS_BLIZZARD[standingID] or "|cfff433ff"
		tooltipColors.standingIncremential = FACTION_COLORS_INC[standingID] or "|cfff433ff"
		tooltipColors.standingIncrementialOld = FACTION_COLORS_INC_old[standingID] or "|cfff433ff"
		
		if not isHeader or hasRep then
			curLine, curColumn = tooltip:AddLine(), 1
			curLine, curColumn = tooltip:SetCell(curLine, curColumn, isWatched and "|TInterface\\Buttons\\UI-CheckBox-Check:14:14|t " or "")
			curLine, curColumn = tooltip:SetCell(curLine, curColumn, tooltipColors[profile.tooltipColorName]                 .. indentation..name, nil, nil, 2)
			curLine, curColumn = tooltip:SetCell(curLine, curColumn, tooltipColors[profile.tooltipColorStandingText]         .. standingText..(hasBonusRepGain and " (+)" or ""))
			curLine, curColumn = tooltip:SetCell(curLine, curColumn, tooltipColors[profile.tooltipColorStandingValueAbsolute].. string_format("%5d / %5d", repIs, repMax))
			curLine, curColumn = tooltip:SetCell(curLine, curColumn, tooltipColors[profile.tooltipColorStandingValuePercent] .. string_format("%.1f%%", repIs / repMax * 100))
			curLine, curColumn = tooltip:SetCell(curLine, curColumn, tooltipColors[profile.tooltipColorStandingValueDelta]   .. GetFactionDeltaText(repDelta))
		else
			curLine, curColumn = tooltip:AddLine(), 2
			curLine, curColumn = tooltip:SetCell(curLine, curColumn, "|cffffffff"..indentation..name, nil, nil, 2)
		end
		tooltip:SetLineScript(curLine, "OnMouseUp", tooltip_line_OnMouseUp, factionIndex)
	end
	
	if self.db.profile.tooltipShowHint then
		tooltip:AddSeparator()
		curLine, curColumn = tooltip:AddLine(), 1
		curLine, curColumn = tooltip:SetCell(curLine, curColumn, "|cfffed100Click:", nil, "LEFT", 2)
		curLine, curColumn = tooltip:SetCell(curLine, curColumn, "|cfffed100set watched faction", nil, nil, 5)
		-- curLine, curColumn = tooltip:AddLine(), 1
		-- curLine, curColumn = tooltip:SetCell(curLine, curColumn, "|cfffed100Alt-Click:", nil, "LEFT", 2)
		-- curLine, curColumn = tooltip:SetCell(curLine, curColumn, "|cfffed100set auto-watched faction", nil, nil, 5)
		curLine, curColumn = tooltip:AddLine(), 1
		curLine, curColumn = tooltip:SetCell(curLine, curColumn, "|cfffed100Ctrl-Click (faction):", nil, "LEFT", 2)
		curLine, curColumn = tooltip:SetCell(curLine, curColumn, "|cfffed100set active/inactive", nil, nil, 5)
		curLine, curColumn = tooltip:AddLine(), 1
		curLine, curColumn = tooltip:SetCell(curLine, curColumn, "|cfffed100Ctrl-Click (header):", nil, "LEFT", 2)
		curLine, curColumn = tooltip:SetCell(curLine, curColumn, "|cfffed100collapse/expand", nil, nil, 5)
		curLine, curColumn = tooltip:AddLine(), 1
		curLine, curColumn = tooltip:SetCell(curLine, curColumn, "|cfffed100Shift-Click:", nil, "LEFT", 2)
		curLine, curColumn = tooltip:SetCell(curLine, curColumn, "|cfffed100write to chatbox", nil, nil, 5)
	end
	
	tooltip:UpdateScrolling()
end

function ElkFactions:GetFactionText(pattern, factionIndex, useColor)
	local name, description, standingID, barMin, barMax, barValue, atWarWith, canToggleAtWar, isHeader, isCollapsed, hasRep, isWatched, isChild, factionID, hasBonusRepGain, canBeLFGBonus = GetFactionInfo(factionIndex)
	local standingText = GetText("FACTION_STANDING_LABEL"..standingID, gender)

	-- check if this is a friendship faction 
	local friendID, friendRep, friendMaxRep, friendName, friendText, friendTexture, friendTextLevel, friendThreshold, nextFriendThreshold = GetFriendshipReputation(factionID)
	if (friendID ~= nil) then
		standingText = friendTextLevel
		if ( nextFriendThreshold ) then
			barMin, barMax, barValue = friendThreshold, nextFriendThreshold, friendRep
		else
			barMin, barMax, barValue = 0, 8400, 8400
		end
	end

	local repIs = barValue - barMin
	local repMax = barMax - barMin
	local repDelta = not friendID and reputationSessionBase[factionID] and (barValue - reputationSessionBase[factionID]) or 0 --TODO: fix for friends
	
	pattern = pattern or ""
	
	--[[
		{n}    name
		{s}    standing
		{rc}   current reputation for this level
		{rm}   maximum reputation for this level
		{rp}   reputation percentage for this level
		{rd}   reputation change this session
		{cw}   color: at war
		{cb}   color: Blizzard
		{ci}   color: incremential
		{cd}   color: reputation change this session
		{cr}   color: reset
	]]
	local color_atwar = atWarWith and "|cffff0000" or canToggleAtWar and "|cffffff00" or "|cff00ff00"
	pattern = string_gsub(pattern, "%b{}", {
		["{n}"] = name,
		["{s}"] = standingText,
		["{rc}"] = repIs,
		["{rm}"] = repMax,
		["{rp}"] = string_format("%.1f%%", repIs / repMax * 100),
		["{rd}"] = GetFactionDeltaText(repDelta),
		["{cw}"] = useColor and color_atwar or "",
		["{cb}"] = useColor and (FACTION_COLORS_BLIZZARD[standingID] or "|cfff433ff") or "",
		["{ci}"] = useColor and (FACTION_COLORS_INC[standingID] or "|cfff433ff") or "",
		["{co}"] = useColor and (FACTION_COLORS_INC_old[standingID] or "|cfff433ff") or "",
		["{cd}"] = useColor and GetFactionDeltaColor(repDelta) or "",
		["{cr}"] = useColor and "|r" or "",
	})
	
	return pattern
end

function ElkFactions:GetAutoZoneFaction()
	local currentZone = GetRealZoneText()
	local currentSubZone = GetSubZoneText()
	
	return self.db.profile.AutoZoneFaction[currentZone.." - "..currentSubZone] or self.db.profile.AutoZoneFaction[currentZone]
end

function ElkFactions:ApplyAutoZoneFaction()
	if not self.db.profile.autoZoneFactionEnabled then return end
	
	local autoZoneFaction = self:GetAutoZoneFaction()
	-- is there a auto-faction set for the current zone?
	if autoZoneFaction then
		-- if lastFaction is set, last zone had autofaction, too
		-- if not we have to store the current watched faction
		if self.db.char.lastFaction == nil then
			self.db.char.lastFaction = GetWatchedFactionInfo() or false
		end
		self.SettingAutoFaction = true
		SetWatchedFactionName(autoZoneFaction)
	-- no auto-faction set so restore last faction if set
	elseif self.db.char.lastFaction ~= nil then
		self.SettingAutoFaction = true
		SetWatchedFactionName(self.db.char.lastFaction)
		self.db.char.lastFaction = nil
	end
end

function ElkFactions:Hook_SetWatchedFactionIndex(factionIndex)

--[===[@debug@
	self:Print("SetWatchedFactionIndex("..tostring(factionIndex)..") called...")
--@end-debug@]===]

	-- store index for lookup
	watchedFactionIndex = factionIndex and factionIndex > 0 and factionIndex or nil

	-- if we aren't setting the auto-faction
	if self.SettingAutoFaction then
		self.SettingAutoFaction = nil
	else
		self.db.char.lastFaction = factionIndex and factionIndex > 0 and GetFactionInfo(factionIndex) or nil
		self:ApplyAutoZoneFaction()
	end
end
